# python3.5+ # pip install uvloop aiohttp.

import asyncio
from asyncio import (Queue, Task, ensure_future, gather, get_event_loop,
                     iscoroutine, set_event_loop_policy)
from asyncio import sleep as asyncio_sleep
from functools import wraps
from time import sleep as time_sleep
from time import time as time_time
from urllib.parse import urlparse

import aiohttp

from ._py3_patch import NewResponse, _py36_all_task_patch
from .configs import Config
from .exceptions import FailureException
from .main import Error, NewFuture, Pool, ProcessPool

try:
    import uvloop

    set_event_loop_policy(uvloop.EventLoopPolicy())
except ImportError:
    Config.dummy_logger.debug("Not found uvloop, using default_event_loop.")

__all__ = "NewTask Loop Asyncme coros get_results_generator Frequency Requests".split(
    " ")


class NotSet(object):
    __slots__ = ()

    def __bool__(self):
        return False

    def __nonzero__(self):
        return False


class NewTask(Task):
    """Add some special method & attribute for asyncio.Task.

    Params:
        :param coro: a standard asyncio await in coroutines.

    Attrs:
        :attr cx: blocking until the task finish and return the callback_result.
        :attr x: blocking until the task finish and return the value as `coro` returned.
        :attr task_start_time: timestamp when the task start up.
        :attr task_end_time: timestamp when the task end up.
        :attr task_cost_time: seconds of task costs.
    """

    _PENDING = "PENDING"
    _CANCELLED = "CANCELLED"
    _FINISHED = "FINISHED"
    _RESPONSE_ARGS = ("encoding", "request_encoding", "content")

    def __init__(self, coro, *, loop=None, callback=None, extra_args=None):
        assert iscoroutine(coro), repr(coro)
        super().__init__(coro, loop=loop)
        self._callback_result = NotSet
        self.extra_args = extra_args or ()
        self.task_start_time = time_time()
        self.task_end_time = 0
        self.task_cost_time = 0
        if callback:
            if not isinstance(callback, (list, tuple, set)):
                callback = [callback]
            self.add_done_callback(self.set_task_time)
            for fn in callback:
                # custom callback will update the _callback_result
                self.add_done_callback(self.wrap_callback(fn))

    @staticmethod
    def wrap_callback(function):
        """Set the callback's result as self._callback_result."""

        @wraps(function)
        def wrapped(task):
            task._callback_result = function(task)
            return task._callback_result

        return wrapped

    @staticmethod
    def set_task_time(task):
        task.task_end_time = time_time()
        task.task_cost_time = task.task_end_time - task.task_start_time

    @property
    def _done_callbacks(self):
        """Keep same api for NewFuture."""
        return self._callbacks

    @property
    def cx(self):
        """Return self.callback_result"""
        return self.callback_result

    @property
    def callback_result(self):
        """Blocking until the task finish and return the callback_result.until"""
        if self._state == self._PENDING:
            self._loop.run_until_complete(self)
        if self._callback_result is NotSet:
            result = self.result()
        else:
            result = self._callback_result
        return result

    @property
    def x(self):
        """Blocking until the task finish and return the self.result()"""
        if self._state == self._PENDING:
            self._loop.run_until_complete(self)
        return self.result()

    def __getattr__(self, name):
        return getattr(self.x, name)

    def __setattr__(self, name, value):
        if name in self._RESPONSE_ARGS:
            self.x.__setattr__(name, value)
        else:
            object.__setattr__(self, name, value)


class Loop:
    """Handle the event loop like a thread pool."""

    def __init__(self,
                 n=None,
                 interval=0,
                 timeout=None,
                 default_callback=None,
                 loop=None,
                 **kwargs):
        self._loop = loop
        self.default_callback = default_callback
        self.async_running = False
        self.n = n
        self.interval = interval
        self._timeout = timeout
        self.frequency = Frequency(self.n, self.interval, loop=self.loop)

    @property
    def loop(self):
        return self._loop or get_event_loop()

    def _wrap_coro_function_with_sem(self, coro_func):
        """Decorator set the coro_function has n/interval control."""

        @wraps(coro_func)
        async def new_coro_func(*args, **kwargs):
            if self.frequency:
                async with self.frequency:
                    result = await coro_func(*args, **kwargs)
                    return result
            else:
                result = await coro_func(*args, **kwargs)
                return result

        return new_coro_func

    def run_in_executor(self, executor=None, func=None, *args):
        """If `kwargs` needed, try like this: func=lambda: foo(*args, **kwargs)"""
        return self.loop.run_in_executor(executor, func, *args)

    def run_in_thread_pool(self, pool_size=None, func=None, *args):
        """If `kwargs` needed, try like this: func=lambda: foo(*args, **kwargs)"""
        executor = Pool(pool_size)
        return self.loop.run_in_executor(executor, func, *args)

    def run_in_process_pool(self, pool_size=None, func=None, *args):
        """If `kwargs` needed, try like this: func=lambda: foo(*args, **kwargs)"""
        executor = ProcessPool(pool_size)
        return self.loop.run_in_executor(executor, func, *args)

    def run_coroutine_threadsafe(self, coro, loop=None, callback=None):
        """Be used when loop running in a single non-main thread."""
        if not iscoroutine(coro):
            raise TypeError("A await in coroutines. object is required")
        loop = loop or self.loop
        future = NewFuture(callback=callback)

        def callback_func():
            try:
                asyncio.futures._chain_future(NewTask(coro, loop=loop), future)
            except Exception as exc:
                if future.set_running_or_notify_cancel():
                    future.set_exception(exc)
                raise

        loop.call_soon_threadsafe(callback_func)
        return future

    def apply(self, coro_function, args=None, kwargs=None, callback=None):
        """Submit a coro_function(*args, **kwargs) as NewTask to self.loop with loop.frequncy control.

        ::

            from torequests.dummy import Loop
            import asyncio
            loop = Loop()


            async def test(i):
                result = await asyncio.sleep(1)
                return (loop.frequency, i)


            task = loop.apply(test, [1])
            print(task)
            # loop.x can be ignore
            print(task.x)
            # <NewTask pending coro=<test() running at e:\github\torequests\torequests\dummy.py:154>>
            # (Frequency(None / None, pending: None, interval: 0s), 1)

        """
        args = args or ()
        kwargs = kwargs or {}
        coro = self._wrap_coro_function_with_sem(coro_function)(*args, **kwargs)
        return self.submit(coro, callback=callback)

    def submit(self, coro, callback=None):
        """Submit a coro as NewTask to self.loop without loop.frequncy control.

        ::

            from torequests.dummy import Loop
            import asyncio
            loop = Loop()


            async def test(i):
                result = await asyncio.sleep(1)
                return (loop.frequency, i)


            coro = test(0)
            task = loop.submit(coro)
            print(task)
            # loop.x can be ignore
            loop.x
            print(task.x)
            # <NewTask pending coro=<test() running at e:\github\torequests\temp_code2.py:6>>
            # (Frequency(None / None, pending: None, interval: 0s), 0)

        """
        callback = callback or self.default_callback
        if self.async_running:
            return self.run_coroutine_threadsafe(coro, callback=callback)
        else:
            return NewTask(coro, loop=self.loop, callback=callback)

    def submitter(self, f):
        """Decorator to submit a coro-function as NewTask to self.loop with control.
        Use default_callback frequency of loop."""
        f = self._wrap_coro_function_with_sem(f)

        @wraps(f)
        def wrapped(*args, **kwargs):
            return self.submit(f(*args, **kwargs))

        return wrapped

    @property
    def x(self):
        """return self.run()"""
        return self.run()

    @property
    def todo_tasks(self):
        """Return tasks in loop which its state is pending."""
        tasks = [
            task for task in self.all_tasks if task._state == NewTask._PENDING
        ]
        return tasks

    @property
    def done_tasks(self):
        """Return tasks in loop which its state is not pending."""
        tasks = [
            task for task in self.all_tasks if task._state != NewTask._PENDING
        ]
        return tasks

    def run(self, tasks=None, timeout=NotSet):
        """Block, run loop until all tasks completed."""
        timeout = self._timeout if timeout is NotSet else timeout
        if self.async_running or self.loop.is_running():
            return self.wait_all_tasks_done(timeout)
        else:
            tasks = tasks or self.todo_tasks
            return self.loop.run_until_complete(gather(*tasks, loop=self.loop))

    def wait_all_tasks_done(self, timeout=NotSet, delay=0.5, interval=0.1):
        """Block, only be used while loop running in a single non-main thread."""
        timeout = self._timeout if timeout is NotSet else timeout
        timeout = timeout or float("inf")
        start_time = time_time()
        time_sleep(delay)
        while 1:
            if not self.todo_tasks:
                return self.all_tasks
            if time_time() - start_time > timeout:
                return self.done_tasks
            time_sleep(interval)

    def close(self):
        """Close the event loop."""
        self.loop.close()

    @property
    def all_tasks(self):
        """Return all tasks of the current loop."""
        return _py36_all_task_patch(loop=self.loop)

    async def pendings(self, tasks=None):
        """Used for await in coroutines.
        `await loop.pendings()`
        `await loop.pendings(tasks)`
        """
        tasks = tasks or self.todo_tasks
        await gather(*tasks, loop=self.loop)


def Asyncme(func, n=None, interval=0, default_callback=None, loop=None):
    """Wrap coro_function into the function return NewTask."""
    return coros(n, interval, default_callback, loop)(func)


def coros(n=None, interval=0, default_callback=None, loop=None):
    """Decorator for wrap coro_function into the function return NewTask."""
    submitter = Loop(
        n=n, interval=interval, default_callback=default_callback,
        loop=loop).submitter

    return submitter


def get_results_generator(*args):
    """TODO"""
    raise NotImplementedError


class _mock_sem:
    _value = 0

    async def acquire(self):
        pass

    def release(self):
        pass

    def __bool__(self):
        return False

    def __nonzero__(self):
        return False


class Frequency(object):
    """Frequency controller, means concurrent running n tasks every interval seconds."""
    __slots__ = ("n", "interval", "loop", "q")

    def __init__(self, n=None, interval=0, loop=None):
        self.n = n
        self.interval = interval
        self.loop = loop
        if self.n:
            self.q = Queue(self.n, loop=self.loop)
            for _ in range(n):
                self.q.put_nowait(1)
        else:
            self.q = None

    @classmethod
    def ensure_frequency(cls, frequency, loop=None):
        if isinstance(frequency, cls):
            return frequency
        elif isinstance(frequency, dict):
            return cls(loop=loop, **frequency)
        else:
            return cls(*frequency, loop=loop)

    async def wait_put(self):
        await asyncio_sleep(self.interval)
        await self.q.put(1)
        self.q.task_done()

    async def __aenter__(self):
        if self.q:
            await self.q.get()

    async def __aexit__(self, *args):
        if self.q:
            ensure_future(self.wait_put())

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "Frequency(%s / %s, pending: %s, interval: %ss)" % (
            self.n - self.q.qsize() if self.q else None,
            self.n,
            len(self.q._getters) if self.q else None,
            self.interval,
        )

    def __bool__(self):
        return bool(self.q)


class Requests(Loop):
    """Wrap the aiohttp with NewTask.

    :param n: sometimes the performance is limited by too large "n",
            or raise ValueError: too many file descriptors on select() (win32),
            so n=100 by default.
    :param interval: sleep after each task done if n.
    :param session: special aiohttp.ClientSession.
    :param catch_exception: whether catch and return the Exception instead of raising it.
    :param default_callback: None
    :param frequencies: None or {host: Frequency obj} or {host: [n, interval]}
    :param default_host_frequency: None
    :param kwargs: will used for aiohttp.ClientSession.

    ::

        from torequests.dummy import Requests
        from torequests.logs import print_info
        trequests = Requests(frequencies={'p.3.cn': (2, 2)})
        ss = [
            trequests.get(
                'http://p.3.cn', retry=1, timeout=5,
                callback=lambda x: (len(x.content), print_info(trequests.frequencies)))
            for i in range(4)
        ]
        trequests.x
        ss = [i.cx for i in ss]
        print_info(ss)
        # [2020-02-11 10:46:18] temp_code2.py(7): {'p.3.cn': Frequency(2 / 2, pending: 2, interval: 2s)}
        # [2020-02-11 10:46:18] temp_code2.py(7): {'p.3.cn': Frequency(2 / 2, pending: 2, interval: 2s)}
        # [2020-02-11 10:46:20] temp_code2.py(7): {'p.3.cn': Frequency(2 / 2, pending: 0, interval: 2s)}
        # [2020-02-11 10:46:20] temp_code2.py(7): {'p.3.cn': Frequency(2 / 2, pending: 0, interval: 2s)}
        # [2020-02-11 10:46:20] temp_code2.py(12): [(612, None), (612, None), (612, None), (612, None)]
    """

    def __init__(self,
                 n=None,
                 interval=0,
                 session=NotSet,
                 catch_exception=True,
                 default_callback=None,
                 frequencies=None,
                 default_host_frequency=None,
                 *,
                 loop=None,
                 return_exceptions=NotSet,
                 **kwargs):
        super().__init__(
            loop=loop,
            default_callback=default_callback,
        )
        # Requests object use its own frequency control, instead of the parent class's.
        self.n = n
        self.interval = interval
        # be compatible with old version's arg `return_exceptions`
        self.catch_exception = (catch_exception if return_exceptions is NotSet
                                else return_exceptions)
        self.default_host_frequency = default_host_frequency
        if self.default_host_frequency:
            assert isinstance(self.default_host_frequency, (list, tuple))
        self.global_frequency = Frequency(self.n, self.interval, loop=self.loop)
        self.frequencies = self.ensure_frequencies(frequencies)
        self.session_kwargs = kwargs
        self._closed = False
        self._session = session
        if self._session is not NotSet:
            session._loop = self.loop
            self._session = session
            if self.n:
                self._session.connector._limit = self.n

    async def _ensure_session(self):
        """ensure the same loop"""
        if self._session is NotSet:
            self._session = aiohttp.ClientSession(
                loop=self.loop, **self.session_kwargs)
            if self.n:
                self._session.connector._limit = self.n
        return self._session

    @property
    def session(self):
        return self._ensure_session()

    def ensure_frequencies(self, frequencies):
        """Ensure frequencies is dict of host-frequencies."""
        if not frequencies:
            return {}
        if not isinstance(frequencies, dict):
            raise ValueError("frequencies should be dict")
        frequencies = {
            host: Frequency.ensure_frequency(frequencies[host], loop=self.loop)
            for host in frequencies
        }
        return frequencies

    def set_frequency(self, host, n=None, interval=NotSet):
        """Set frequency for host with n and interval."""
        frequency = Frequency(
            n or self.n,
            self.interval if interval is NotSet else interval,
            loop=self.loop)
        self.update_frequency({host: frequency})
        return frequency

    def update_frequency(self, frequencies):
        """Update the frequencies with dict of new frequencies."""
        self.frequencies.update(self.ensure_frequencies(frequencies))

    async def _request(self, method, url, retry=0, **kwargs):
        url = url.strip()
        parsed_url = urlparse(url)
        scheme = parsed_url.scheme
        host = parsed_url.netloc
        if host in self.frequencies:
            frequency = self.frequencies[host]
        elif self.default_host_frequency:
            frequency = self.set_frequency(host, *self.default_host_frequency)
        else:
            frequency = self.global_frequency
        if 'timeout' in kwargs:
            # for timeout=(1,2) and timeout=5
            if isinstance(kwargs['timeout'], (tuple, list)):
                kwargs['timeout'] = aiohttp.client.ClientTimeout(
                    sock_connect=kwargs['timeout'][0],
                    sock_read=kwargs['timeout'][1])
            elif isinstance(kwargs['timeout'], aiohttp.client.ClientTimeout):
                pass
            else:
                kwargs['timeout'] = aiohttp.client.ClientTimeout(
                    sock_connect=kwargs['timeout'], sock_read=kwargs['timeout'])
        proxies = kwargs.pop("proxies", None)
        if "verify" in kwargs:
            kwargs["verify_ssl"] = kwargs.pop("verify")
        if proxies:
            kwargs["proxy"] = "%s://%s" % (scheme, proxies[scheme])
        kwargs["url"] = url
        kwargs["method"] = method
        # non-official request args
        referer_info = kwargs.pop("referer_info", None)
        encoding = kwargs.pop("encoding", None)
        for retries in range(retry + 1):
            async with frequency:
                try:
                    session = await self.session
                    async with session.request(**kwargs) as resp:
                        await resp.read()
                        r = NewResponse(resp, encoding=encoding)
                        r.referer_info = referer_info
                        return r
                except (aiohttp.ClientError, Error) as err:
                    error = err
                    continue
        else:
            kwargs["retry"] = retry
            if referer_info:
                kwargs["referer_info"] = referer_info
            if encoding:
                kwargs["encoding"] = encoding
            error.request = kwargs
            Config.dummy_logger.debug("Retry %s & failed: %s." % (retry, error))
            if self.catch_exception:
                failure = FailureException(error)
                failure.request = kwargs
                return failure
            raise error

    def request(self, method, url, callback=None, retry=0, **kwargs):
        """Submit the coro of self._request to self.loop"""
        return self.submit(
            self._request(method, url=url, retry=retry, **kwargs),
            callback=(callback or self.default_callback),
        )

    def get(self, url, params=None, callback=None, retry=0, **kwargs):
        return self.request(
            "get",
            url=url,
            params=params,
            callback=callback,
            retry=retry,
            **kwargs)

    def post(self, url, data=None, callback=None, retry=0, **kwargs):
        return self.request(
            "post",
            url=url,
            data=data,
            callback=callback,
            retry=retry,
            **kwargs)

    def delete(self, url, callback=None, retry=0, **kwargs):
        return self.request(
            "delete", url=url, callback=callback, retry=retry, **kwargs)

    def put(self, url, data=None, callback=None, retry=0, **kwargs):
        return self.request(
            "put", url=url, data=data, callback=callback, retry=retry, **kwargs)

    def head(self, url, callback=None, retry=0, **kwargs):
        return self.request(
            "head", url=url, callback=callback, retry=retry, **kwargs)

    def options(self, url, callback=None, retry=0, **kwargs):
        return self.request(
            "options", url=url, callback=callback, retry=retry, **kwargs)

    def patch(self, url, callback=None, retry=0, **kwargs):
        return self.request(
            "patch", url=url, callback=callback, retry=retry, **kwargs)

    async def close(self):
        if self._closed:
            return
        try:
            session = await self._ensure_session()
            if session and not session.closed:
                await session.close()
            self._closed = True
        except Exception as e:
            Config.dummy_logger.error("can not close session for: %s" % e)

    def __del__(self):
        _exhaust_simple_coro(self.close())

    def __enter__(self):
        return self

    def __exit__(self, *args):
        _exhaust_simple_coro(self.close())

    async def __aenter__(self):
        await self.session
        return self

    async def __aexit__(self, *args):
        await self.close()


def _exhaust_simple_coro(coro):
    """Run coroutines without event loop, only support simple coroutines which can run without future.
    Or it will raise RuntimeError: await wasn't used with future."""
    while True:
        try:
            coro.send(None)
        except StopIteration as e:
            return e.value
